from struct import unpack

from pydicom.dataelem import DataElement
from pydicom.filebase import DicomBytesIO
from pydicom.filereader import read_partial, read_sequence_item
from pydicom.filewriter import write_sequence_item, write_dataset
from pydicom.sequence import Sequence
from pydicom.tag import SequenceDelimiterTag, Tag

from pydicom.dataset import Dataset
from pydicom.charset import default_encoding


def encode(ds, is_implicit_vr, is_little_endian, is_sequence_item=False):
    """Encode a pydicom Dataset `ds` to a byte stream.
    (from https://github.com/pydicom/pynetdicom3)
    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset to encode
    is_implicit_vr : bool
        The element encoding scheme the dataset will be encoded with.
    is_little_endian : bool
        The byte ordering the dataset will be encoded in.
    is_sequence_item: bool
        Set to true to write a sequence item

    Returns
    -------
    bytes or None
        The encoded dataset (if successful)
    """
    # pylint: disable=broad-except
    fp = DicomBytesIO()
    fp.is_implicit_VR = is_implicit_vr
    fp.is_little_endian = is_little_endian
    try:
        if is_sequence_item:
            write_sequence_item(fp, ds, default_encoding)
        else:
            write_dataset(fp, ds)
    except Exception:
        fp.close()
        raise

    bytestring = fp.parent.getvalue()
    fp.close()

    return bytestring


class RawWrapper(object):

    def __init__(self, input_stream, wrapped_content_tag, is_implicit_vr=False, is_little_endian=True, buffer_size=2**20):
        self._input_stream = input_stream
        self._is_implicit_vr = is_implicit_vr
        self._is_little_endian = is_little_endian
        self._buffer_size = buffer_size
        if buffer_size % 2 == 1:
            raise RuntimeError("DICOM standard mandates even-sized value length, so buffer_size must be even.")
        dataset = Dataset()
        dataset[wrapped_content_tag] = DataElement(wrapped_content_tag, "SQ", Sequence(), is_undefined_length=True)
        encoded = encode(dataset, is_implicit_vr=is_implicit_vr, is_little_endian=is_little_endian)

        # a low-level method of splitting the empty dataset into prefix
        # (DicomWrapperCreator element and DicomWrapperContent SQ element header) and suffix (end of sequence delimiter)
        length_of_sequence_delimiter = 8
        self._prefix = encoded[:-length_of_sequence_delimiter]
        self._suffix = encoded[-length_of_sequence_delimiter:]

    def _encapsulate(self, input_data):
        first_sequence_element = self._prefix is not None

        embedded_ds = Dataset()
        embedded_ds.EncapsulatedDocument = input_data
        encapsulated_sequence_element = encode(
            embedded_ds,
            is_implicit_vr=self._is_implicit_vr,
            is_little_endian=self._is_little_endian,
            is_sequence_item=True
        )

        # prepend prefix is first sequence element
        if first_sequence_element:
            encapsulated_sequence_element = self._prefix + encapsulated_sequence_element
            self._prefix = None
        return encapsulated_sequence_element

    def __iter__(self):

        # cache one data chunk to keep track of which is the last one, in order to append padding byte(s) to it
        input_data = self._input_stream.read(self._buffer_size)
        while True:
            next_data = self._input_stream.read(self._buffer_size)
            incoming_data_length = len(next_data)
            if incoming_data_length == 0:
                # length of next_data is 0, therefore input_data is the last data chunk
                # append padding byte(s)
                if len(input_data) % 2:
                    input_data += "\01"
                else:
                    input_data += "\00\02"
                yield self._encapsulate(input_data)
                yield self._suffix
                return

            yield self._encapsulate(input_data)
            input_data = next_data


class RawUnwrapper(object):

    def __init__(self, input_stream, wrapped_content_tag, is_implicit_vr=False, is_little_endian=True):
        self._is_implicit_vr = is_implicit_vr
        self._is_little_endian = is_little_endian
        self._input_stream = input_stream
        self._wrapped_content_tag = wrapped_content_tag

    def _wrapped_sequence_start(self, tag, *_):
        return tag == self._wrapped_content_tag

    def _endian_format(self):
        if self._is_little_endian:
            return '<'
        else:
            return '>'

    def _end_of_sequence(self):
        # look ahead to next tag and 4 bytes after it
        next_8bytes = self._input_stream.read(8)
        # rewind
        self._input_stream.seek(self._input_stream.tell() - 8)
        next_group, next_element, next_4bytes = unpack(self._endian_format() + "HHL", next_8bytes)

        next_tag = Tag(next_group, next_element)
        return next_tag == SequenceDelimiterTag and next_4bytes == 0

    def __iter__(self):
        # discard dataset elements prior to the wrapped sequence tag
        read_partial(self._input_stream, stop_when=self._wrapped_sequence_start)
        # offset 12 bytes of sequence tag, VR, length in order to advance the buffer to position of first sequence item
        sequence_header_length = 8 if self._is_implicit_vr else 12

        # cache one data chunk to keep track of which is the last one, in order to remove padding byte(s) from it
        self._input_stream.read(sequence_header_length)
        embedded_ds = read_sequence_item(
            self._input_stream, encoding=default_encoding, is_implicit_VR=self._is_implicit_vr,
            is_little_endian=self._is_little_endian)
        while True:
            data = embedded_ds.EncapsulatedDocument
            # explicitly test whether at end of file
            if self._input_stream.read(1) == '':
                next_ds = None
            else:
                self._input_stream.seek(-1, 1)
                next_ds = read_sequence_item(
                    self._input_stream, encoding=default_encoding, is_implicit_VR=self._is_implicit_vr,
                    is_little_endian=self._is_little_endian)

            if next_ds:
                yield data
                embedded_ds = next_ds
            else:
                # remove padding from last data chunk
                num_padding_bytes = ord(data[-1])
                data = data[:-num_padding_bytes]
                yield data
                return


class Wrapper(RawWrapper):

    def __init__(self, input_stream, private_group, content_creator,
                 is_implicit_vr=False, is_little_endian=True, buffer_size=2**20):

        super(Wrapper, self).__init__(input_stream,
                                      wrapped_content_tag=Tag(private_group, 0x1000),
                                      is_implicit_vr=is_implicit_vr,
                                      is_little_endian=is_little_endian,
                                      buffer_size=buffer_size
                                      )
        dataset = Dataset()
        content_creator_tag = Tag(private_group, 0x0010)
        dataset[content_creator_tag] = DataElement(content_creator_tag, "LO", content_creator)
        encoded = encode(dataset, is_implicit_vr=is_implicit_vr, is_little_endian=is_little_endian)
        self._prefix = encoded + self._prefix


class Unwrapper(RawUnwrapper):
    def __init__(self, input_stream, private_group, is_implicit_vr=False, is_little_endian=True):
        super(Unwrapper, self).__init__(input_stream,
                                        wrapped_content_tag=Tag(private_group, 0x1000),
                                        is_implicit_vr=is_implicit_vr,
                                        is_little_endian=is_little_endian)



import glob
import os
import types

import zipstream
from zipfile import ZIP_DEFLATED


class IterStreamer(object):
    """
    Allows file-like read operations (read, seek, tell) on iterator contents. Seek and tell are subject to buffer size.
    """
    def __init__(self, iterable, buffer_size=2 ** 20):
        """
        Initialize.

        :param iterable: the iterable object to be streamed
        :param buffer_size: a guideline length for the internal buffer. The buffer is of dynamic length, and buffer_size
        is used only for guiding the trimming. The Buffer is expanded by consuming from the iterable. It is trimmed to
        buffer_size/2 when length of the portion of the buffer before current position exceeds buffer_size/2. Trimming
        is invoked by read().

        """
        self._buffer_size = buffer_size
        self._iterable = iter(iterable)
        self._global_position = 0
        self._position_in_buffer = 0
        self._buffer = ''
        self._end_of_iterator = False

    def _trim_buffer(self):
        # see comment on parameter buffer_size of __init__ above
        if self._position_in_buffer > self._buffer_size/2:
            new_start = self._position_in_buffer - self._buffer_size/2
            self._buffer = self._buffer[new_start:]
            self._position_in_buffer -= new_start

    def read(self, size):
        length_to_return = (len(self._buffer) - self._position_in_buffer)
        try:
            while (len(self._buffer) - self._position_in_buffer) < size:
                data = next(self._iterable)
                length_to_return += len(data)
                self._buffer += data
        except StopIteration:
            # actually read read_data_length
            size = length_to_return
        self._trim_buffer()
        # advance the position pointers
        self._position_in_buffer += size
        self._global_position += size
        # since the pointers have been advanced, start and end indices of return value are computed relative to the new
        # position_in_buffer
        return self._buffer[self._position_in_buffer-size:self._position_in_buffer]

    def seek(self, offset, from_what=0):
        if from_what != 0 and from_what != 1:
            raise IOError('Offsets from end are not supported')
        elif from_what == 0:
            self.seek(offset-self._global_position, 1)
        else:
            new_position = self._position_in_buffer+offset
            if new_position < 0 or new_position > len(self._buffer):
                raise IOError('Cannot fast-forward/rewind outside current buffer.')
            self._position_in_buffer = new_position
            self._global_position += offset

    def tell(self):
        return self._global_position


class StreamingArchive(object):
    def __init__(self, input_data, archive_name, compress_type=ZIP_DEFLATED, progress_callback=None):
        """
        StreamingArchive can be used for streaming compression of input_data contents using a file-like stream.

        :param input_data: data to compress. Can be a list of filepaths, a string holding a valid path to either a file
            or a directory, an iterable, or a list of iterables.
        :param archive_name: name for the root directory.
        :param compress_type: either zipfile.ZIP_DEFLATED (compressed with deflate algorithm) or zipfile.ZIP_STORED
            (uncompressed)
        :param progress_callback: optional callback which will be called every ~500kb with a percentage progress.
            This only works properly if all items being compressed are files, not iterables.
        """
        self._progress_callback = progress_callback
        self._zipfile = zipstream.ZipFile(allowZip64=True)
        # if the installed version of zipstream includes the progress bar customization, use it. Otherwise, report 50%
        if hasattr(self._zipfile, 'get_progress'):
            self._zipfile_progress = self._zipfile.get_progress
        else:
            self._zipfile_progress = lambda: 0.5

        def iterable_item_and_arcname(itr):
            return itr, '{0}{1}iterable_{2}'.format(archive_name, os.sep, itr.__hash__())

        if isinstance(input_data, str) or isinstance(input_data, unicode):
            # a string input_data needs to be a valid file or directory
            if not os.path.exists(input_data) or (not os.path.isfile(input_data) and not os.path.isdir(input_data)):
                raise IOError(
                        "Invalid input parameter input_files: {0} is a string but not a valid file/directory path".format(input_data)
                    )
            if os.path.isfile(input_data):
                # items needs to be a list, make a list of one element
                items = [(
                    input_data,
                    '{0}{1}{2}'.format(archive_name, os.sep, os.path.basename(input_data))
                )]
            else:
                # this is a directory. compress all files in the directory.
                items = []
                for root, _, files in os.walk(input_data):
                    for name in files:
                        fpath = root + os.sep + name
                        # for arcname, use path relative to the root directory (passed as input_data)
                        items.append((
                            fpath,
                            '{0}{1}{2}'.format(archive_name, os.sep, os.path.relpath(fpath, start=input_data))
                        ))
        elif isinstance(input_data, types.GeneratorType):
            # items needs to be a list, make a list of one element
            items = [(iterable_item_and_arcname(input_data))]
        elif isinstance(input_data, list):
            # every element in a list needs to be an iterable or a valid file path
            items = []
            for item in input_data:
                if isinstance(item, types.GeneratorType):
                    items.append((iterable_item_and_arcname(item)))
                elif isinstance(item, str) or isinstance(item, unicode):
                    if not os.path.exists(item) or not os.path.isfile(item):
                        raise IOError(
                            "Invalid input list element: {0} is a string but not a valid file path".format(item)
                        )
                    items.append((
                        item,
                        '{0}{1}{2}'.format(archive_name, os.sep, os.path.basename(item))
                    ))
        else:
            raise RuntimeError('input_date has invalid type: must be one of (str, unicode, types.GeneratorType, list)')

        for item, arcname in items:
            if isinstance(item, types.GeneratorType):
                self._zipfile.write_iter(arcname=arcname, iterable=item, compress_type=compress_type)
            elif os.path.exists(item) and os.path.isfile(item):
                self._zipfile.write(filename=item, arcname=arcname, compress_type=compress_type)
            else:
                raise IOError("Invalid list or generator element: {0}. Input_paths == {1}".format(item, input_data))

    def get_progress(self):
        """
        Get percentage of uncompressed data that has already been processed.
        This only works properly if all items being compressed are files, not iterables.
        """
        return self._zipfile_progress()

    def get_compressed_stream(self):
        """
        Get a file-like stream of compressed data
        :return:
        """
        if self._progress_callback and callable(self._progress_callback):
            def report_progress(itr):
                for index, item in enumerate(itr):
                    # zipstream library reads files in 8192 byte increments. This is too fine-grained,
                    # so report progress for every ~500kb
                    if index % 2**6 == 0:
                        self._progress_callback(self._zipfile_progress())
                    yield item
                self._progress_callback(self._zipfile_progress())
            iterable = report_progress(self._zipfile)
        else:
            iterable = self._zipfile
        return IterStreamer(iterable)
